import zendriver as zd

from tests.sample_data import sample_file
from zendriver import SpecialKeys, KeyModifiers, KeyEvents


async def test_visible_events(browser: zd.Browser) -> None:
    """Test keyboard events with contenteditable div."""
    # Open the page
    main_page = await browser.get(sample_file("simple_editor.html"))

    text_part = await main_page.find('//*[@id="editor"]')

    await text_part.mouse_click("left")
    await main_page.sleep(1)  # give some time to focus the text part
    await text_part.send_keys("Hello, world!")

    payloads = KeyEvents.from_mixed_input(
        [
            " This is another sentence",
            SpecialKeys.ENTER,
            ("a", KeyModifiers.Ctrl),
            ("c", KeyModifiers.Ctrl),
            SpecialKeys.ARROW_UP,
            ("v", KeyModifiers.Ctrl),
            " This is pasted text. 👍",
        ]
    )

    await text_part.send_keys(payloads)
    check_part = await main_page.find('//*[@id="editor"]')
    assert len(check_part.children) == 4, "Expected 4 children after operations"

    expected_output = [
        "Hello, world! This is another sentence",
        "<div>&nbsp;This is pasted text. 👍</div>",
        "Hello, world! This is another sentence",
        "<div><br></div>",
    ]

    for expected, actual_text in zip(expected_output, check_part.children):
        actual_html = await actual_text.get_html()
        assert actual_html == expected, f"Expected '{expected}', got '{actual_html}'"


async def test_escape_key_popup(browser: zd.Browser) -> None:
    """Test escape key functionality to close a popup."""
    main_page = await browser.get(sample_file("special_key_detector.html"))

    status_check = await main_page.find('//*[@id="status"]')
    assert (
        status_check.text == "Ready - Click button to open popUp"
    ), "There is something wrong with the page"

    button = await main_page.find('//*[@id="mainpageButton"]')
    await button.mouse_click("left")

    await status_check
    assert (
        status_check.text == "popUp is OPEN - Press Escape to close"
    ), "Popup did not open correctly"

    pop_up = await main_page.find('//*[@id="mainpage"]/div')
    await pop_up.send_keys(SpecialKeys.ESCAPE)

    await status_check
    assert (
        status_check.text == "popUp is CLOSED - Click button to open again"
    ), "Popup did not close correctly"



from zendriver import cdp
from zendriver._version import __version__
from zendriver.core import util
from zendriver.core._contradict import (
    ContraDict,  # noqa
    cdict,
)
from zendriver.core.browser import Browser
from zendriver.core.config import Config
from zendriver.core.connection import Connection
from zendriver.core.element import Element
from zendriver.core.tab import Tab
from zendriver.core.util import loop, start
from zendriver.core.keys import KeyEvents, SpecialKeys, KeyPressEvent, KeyModifiers

__all__ = [
    "__version__",
    "loop",
    "Browser",
    "Tab",
    "cdp",
    "Config",
    "start",
    "util",
    "Element",
    "ContraDict",
    "cdict",
    "Connection",
    "KeyEvents",
    "SpecialKeys",
    "KeyPressEvent",
    "KeyModifiers",
]



import asyncio
from typing import Any

import pytest

import zendriver as zd
from tests.sample_data import sample_file
from zendriver.cdp.fetch import RequestStage
from zendriver.cdp.network import ResourceType
from zendriver.core.connection import ProtocolException


async def test_set_user_agent_sets_navigator_values(browser: zd.Browser) -> None:
    tab = browser.main_tab
    assert tab is not None

    await tab.set_user_agent(
        "Test user agent", accept_language="testLang", platform="TestPlatform"
    )

    navigator_user_agent = await tab.evaluate("navigator.userAgent")
    navigator_language = await tab.evaluate("navigator.language")
    navigator_platform = await tab.evaluate("navigator.platform")
    assert navigator_user_agent == "Test user agent"
    assert navigator_language == "testLang"
    assert navigator_platform == "TestPlatform"


async def test_set_user_agent_defaults_existing_user_agent(browser: zd.Browser) -> None:
    tab = browser.main_tab
    assert tab is not None

    existing_user_agent = await tab.evaluate("navigator.userAgent")

    await tab.set_user_agent(accept_language="testLang")

    navigator_user_agent = await tab.evaluate("navigator.userAgent")
    navigator_language = await tab.evaluate("navigator.language")
    assert navigator_user_agent == existing_user_agent
    assert navigator_language == "testLang"


async def test_find_finds_element_by_text(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("groceries.html"))

    result = await tab.find("Apples")

    assert result is not None
    assert result.tag == "li"
    assert result.text == "Apples"


async def test_find_times_out_if_element_not_found(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("groceries.html"))

    with pytest.raises(asyncio.TimeoutError):
        await tab.find("Clothes", timeout=1)


async def test_select(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("groceries.html"))

    result = await tab.select("li[aria-label^='Apples']")

    assert result is not None
    assert result.tag == "li"
    assert result.text == "Apples"


async def test_xpath(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("groceries.html"))

    results = await tab.xpath('//li[@aria-label="Apples (42)"]')

    assert len(results) == 1
    result = results[0]

    assert result is not None
    assert result.tag == "li"
    assert result.text == "Apples"


async def test_add_handler_type_event(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("groceries.html"))

    async def request_handler_1(event: zd.cdp.network.RequestWillBeSent) -> None:
        pass

    async def request_handler_2(event: zd.cdp.network.RequestWillBeSent) -> None:
        pass

    assert len(tab.handlers) == 0

    tab.add_handler(zd.cdp.network.RequestWillBeSent, request_handler_1)

    tab.add_handler(zd.cdp.network.RequestWillBeSent, request_handler_2)

    assert len(tab.handlers) == 1
    assert len(tab.handlers[zd.cdp.network.RequestWillBeSent]) == 2
    assert tab.handlers[zd.cdp.network.RequestWillBeSent] == [
        request_handler_1,
        request_handler_2,
    ]


async def test_add_handler_module_event(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("groceries.html"))

    async def request_handler(event: Any) -> None:
        pass

    assert len(tab.handlers) == 0

    tab.add_handler(zd.cdp.network, request_handler)

    assert len(tab.handlers) == 27


async def test_remove_handlers(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("groceries.html"))

    async def request_handler(event: zd.cdp.network.RequestWillBeSent) -> None:
        pass

    tab.add_handler(zd.cdp.network.RequestWillBeSent, request_handler)
    assert len(tab.handlers) == 1

    tab.remove_handlers()
    assert len(tab.handlers) == 0


async def test_remove_handlers_specific_event(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("groceries.html"))

    async def request_handler(event: zd.cdp.network.RequestWillBeSent) -> None:
        pass

    tab.add_handler(zd.cdp.network.RequestWillBeSent, request_handler)
    assert len(tab.handlers) == 1

    tab.remove_handlers(
        zd.cdp.network.RequestWillBeSent,
    )
    assert len(tab.handlers) == 0


async def test_remove_specific_handler(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("groceries.html"))

    async def request_handler_1(event: zd.cdp.network.RequestWillBeSent) -> None:
        pass

    async def request_handler_2(event: zd.cdp.network.RequestWillBeSent) -> None:
        pass

    tab.add_handler(zd.cdp.network.RequestWillBeSent, request_handler_1)
    tab.add_handler(zd.cdp.network.RequestWillBeSent, request_handler_2)
    assert len(tab.handlers) == 1
    assert len(tab.handlers[zd.cdp.network.RequestWillBeSent]) == 2

    tab.remove_handlers(zd.cdp.network.RequestWillBeSent, request_handler_1)
    assert len(tab.handlers) == 1
    assert len(tab.handlers[zd.cdp.network.RequestWillBeSent]) == 1


async def test_remove_handlers_without_event(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("groceries.html"))

    async def request_handler(event: zd.cdp.network.RequestWillBeSent) -> None:
        pass

    tab.add_handler(zd.cdp.network.RequestWillBeSent, request_handler)
    assert len(tab.handlers) == 1

    with pytest.raises(ValueError) as e:
        tab.remove_handlers(handler=request_handler)
        assert str(e) == "if handler is provided, event_type should be provided as well"


async def test_wait_for_ready_state(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("groceries.html"))

    await tab.wait_for_ready_state("complete")

    ready_state = await tab.evaluate("document.readyState")
    assert ready_state == "complete"


async def test_expect_request(browser: zd.Browser) -> None:
    tab = browser.main_tab
    assert tab is not None

    async with tab.expect_request(sample_file("groceries.html")) as request_info:
        await tab.get(sample_file("groceries.html"))
        req = await asyncio.wait_for(request_info.value, timeout=3)
        assert type(req) is zd.cdp.network.RequestWillBeSent
        assert type(req.request) is zd.cdp.network.Request
        assert req.request.url == sample_file("groceries.html")
        assert req.request_id is not None

        response_body = await request_info.response_body
        assert response_body is not None
        assert type(response_body) is tuple


async def test_expect_response(browser: zd.Browser) -> None:
    tab = browser.main_tab
    assert tab is not None

    async with tab.expect_response(sample_file("groceries.html")) as response_info:
        await tab.get(sample_file("groceries.html"))
        resp = await asyncio.wait_for(response_info.value, timeout=3)
        assert type(resp) is zd.cdp.network.ResponseReceived
        assert type(resp.response) is zd.cdp.network.Response
        assert resp.request_id is not None

        response_body = await response_info.response_body
        assert response_body is not None
        assert type(response_body) is tuple


async def test_expect_response_with_reload(browser: zd.Browser) -> None:
    tab = browser.main_tab
    assert tab is not None

    async with tab.expect_response(sample_file("groceries.html")) as response_info:
        await tab.get(sample_file("groceries.html"))
        await tab.wait_for_ready_state("complete")
        await response_info.reset()
        await tab.reload()
        await tab.wait_for_ready_state("complete")
        resp = await asyncio.wait_for(response_info.value, timeout=3)
        assert type(resp) is zd.cdp.network.ResponseReceived
        assert type(resp.response) is zd.cdp.network.Response
        assert resp.request_id is not None

        response_body = await response_info.response_body
        assert response_body is not None
        assert type(response_body) is tuple


async def test_expect_download(browser: zd.Browser) -> None:
    tab = browser.main_tab
    assert tab is not None

    async with tab.expect_download() as download_ex:
        await tab.get(sample_file("groceries.html"))
        await (await tab.select("#download_file")).click()
        download = await asyncio.wait_for(download_ex.value, timeout=3)
        assert type(download) is zd.cdp.browser.DownloadWillBegin
        assert download.url is not None


async def test_intercept(browser: zd.Browser) -> None:
    tab = browser.main_tab
    assert tab is not None

    async with tab.intercept(
        "*/user-data.json",
        RequestStage.RESPONSE,
        ResourceType.XHR,
    ) as interception:
        await tab.get(sample_file("profile.html"))
        body, _ = await interception.response_body
        await interception.continue_request()

        assert body is not None
        # original_response = loads(body)
        # assert original_response["name"] == "Zendriver"


async def test_intercept_with_reload(browser: zd.Browser) -> None:
    tab = browser.main_tab
    assert tab is not None

    async with tab.intercept(
        "*/user-data.json",
        RequestStage.RESPONSE,
        ResourceType.XHR,
    ) as interception:
        await tab.get(sample_file("profile.html"))
        await interception.response_body
        await interception.continue_request()

        await interception.reset()
        await tab.reload()
        body, _ = await interception.response_body
        await interception.continue_request()

        assert body is not None
        # original_response = loads(body)
        # assert original_response["name"] == "Zendriver"


async def test_evaluate_complex_object_no_error(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("complex_object.html"))
    await tab.wait_for_ready_state("complete")

    result = await tab.evaluate(
        "document.querySelector('body:not(.no-js)')", return_by_value=False
    )
    assert result is not None

    # This is similar to the original failing case but more likely to trigger the error
    body_with_complex_refs = await tab.evaluate("document.body", return_by_value=False)
    assert body_with_complex_refs is not None


async def test_evaluate_return_by_value_complex_object(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("complex_object.html"))
    await tab.wait_for_ready_state("complete")

    expression = "document.querySelector('body:not(.no-js)')"

    # Fetching a complex object with return_by_value=True is unsupported because there is no
    # way to represent the object as a simple data structure.
    with pytest.raises(ProtocolException):
        _ = await tab.evaluate(expression, return_by_value=True)

    result_by_value_false = await tab.evaluate(expression, return_by_value=False)
    assert (
        result_by_value_false is not None
    )  # Should return the deep serialized value, not a tuple


async def test_evaluate_return_by_value_simple_json(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("simple_json.html"))
    await tab.wait_for_ready_state("complete")

    expression = "JSON.parse(document.querySelector('#obj').textContent)"

    result_by_value_true = await tab.evaluate(expression, return_by_value=True)
    assert result_by_value_true == {"a": "x", "b": 3.14159}

    result_by_value_false = await tab.evaluate(expression, return_by_value=False)
    assert result_by_value_false == [
        ["a", {"type": "string", "value": "x"}],
        ["b", {"type": "number", "value": 3.14159}],
    ]


async def test_evaluate_return_by_value_falsy(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("simple_json.html"))
    await tab.wait_for_ready_state("complete")

    expr_template = "JSON.parse(document.querySelector('%s').textContent)"

    assert await tab.evaluate(expr_template % "#zero") == 0
    assert await tab.evaluate(expr_template % "#empty_array") == []
    assert await tab.evaluate(expr_template % "#null") is None


async def test_evaluate_stress_test_complex_objects(browser: zd.Browser) -> None:
    tab = await browser.get(sample_file("complex_object.html"))
    await tab.wait_for_ready_state("complete")

    # Test various DOM queries that could trigger reference chain issues
    # Each test case is a tuple of (expression, return_by_value, expected_type_or_validator)
    test_cases = [
        ("document.querySelector('body:not(.no-js)')", False, lambda x: x is not None),
        ("document.documentElement", False, lambda x: x is not None),
        ("document.querySelector('*')", False, lambda x: x is not None),
        ("document.body.parentElement", False, lambda x: x is not None),
        ("document.getElementById('content')", False, lambda x: x is not None),
        (
            "document.body.complexStructure ? 'has complex structure' : 'no structure'",
            True,
            str,
        ),
        ("document.readyState", True, str),
        ("navigator.userAgent", True, str),
        ("window.location.href", True, str),
        ("document.title", True, str),
    ]

    for expression, return_by_value, validator in test_cases:
        result = await tab.evaluate(expression, return_by_value=return_by_value)
        # Verify the result is usable and matches expected type/validation
        if callable(validator):
            assert validator(
                result
            ), f"Result validation failed for '{expression}': {result}"
        elif isinstance(validator, type):
            assert isinstance(
                result, validator
            ), f"Expected {validator} for '{expression}', got {type(result)}: {result}"
        else:
            raise ValueError("Validator must be a type or callable")



import asyncio
import typing

from zendriver import cdp
from zendriver.cdp.fetch import HeaderEntry, RequestStage, RequestPattern
from zendriver.cdp.network import ResourceType
from zendriver.core.connection import Connection


class BaseFetchInterception:
    """
    Base class to wait for a Fetch response matching a URL pattern.
    Use this to collect and decode a paused fetch response, while keeping
    the use block clean and returning its own result.

    :param tab: The Tab instance to monitor.
    :param url_pattern: The URL pattern to match requests and responses.
    :param request_stage: The stage of the fetch request to intercept (e.g., request or response).
    :param resource_type: The type of resource to intercept (e.g., document, script, etc.).
    """

    def __init__(
        self,
        tab: Connection,
        url_pattern: str,
        request_stage: RequestStage,
        resource_type: ResourceType,
    ):
        self.tab = tab
        self.url_pattern = url_pattern
        self.request_stage = request_stage
        self.resource_type = resource_type
        self.response_future: asyncio.Future[cdp.fetch.RequestPaused] = asyncio.Future()

    async def _response_handler(self, event: cdp.fetch.RequestPaused) -> None:
        """
        Internal handler for response events.
        :param event: The response event.
        :type event: cdp.fetch.RequestPaused
        """
        self._remove_response_handler()
        self.response_future.set_result(event)

    def _remove_response_handler(self) -> None:
        """
        Remove the response event handler.
        """
        self.tab.remove_handlers(cdp.fetch.RequestPaused, self._response_handler)

    async def __aenter__(self) -> "BaseFetchInterception":
        """
        Enter the context manager, adding request and response handlers.
        """
        await self._setup()
        return self

    async def __aexit__(self, *args: typing.Any) -> None:
        """
        Exit the context manager, removing request and response handlers.
        """
        await self._teardown()

    async def _setup(self) -> None:
        await self.tab.send(
            cdp.fetch.enable(
                [
                    RequestPattern(
                        url_pattern=self.url_pattern,
                        request_stage=self.request_stage,
                        resource_type=self.resource_type,
                    )
                ]
            )
        )
        self.tab.enabled_domains.append(
            cdp.fetch
        )  # trick to avoid another `fetch.enable` call by _register_handlers
        self.tab.add_handler(cdp.fetch.RequestPaused, self._response_handler)

    async def _teardown(self) -> None:
        self._remove_response_handler()
        await self.tab.send(cdp.fetch.disable())

    async def reset(self) -> None:
        """
        Resets the internal state, allowing the interception to be reused.
        """
        self.response_future = asyncio.Future()
        await self._teardown()
        await self._setup()

    @property
    async def request(self) -> cdp.network.Request:
        """
        Get the matched request.
        :return: The matched request.
        :rtype: cdp.network.request
        """
        return (await self.response_future).request

    @property
    async def response_body(self) -> tuple[str, bool]:
        """
        Get the body of the matched response.
        :return: The response body.
        :rtype: str
        """
        request_id = (await self.response_future).request_id
        body = await self.tab.send(cdp.fetch.get_response_body(request_id=request_id))
        return body

    async def fail_request(self, error_reason: cdp.network.ErrorReason) -> None:
        request_id = (await self.response_future).request_id
        await self.tab.send(
            cdp.fetch.fail_request(request_id=request_id, error_reason=error_reason)
        )

    async def continue_request(
        self,
        url: typing.Optional[str] = None,
        method: typing.Optional[str] = None,
        post_data: typing.Optional[str] = None,
        headers: typing.Optional[typing.List[HeaderEntry]] = None,
        intercept_response: typing.Optional[bool] = None,
    ) -> None:
        request_id = (await self.response_future).request_id
        await self.tab.send(
            cdp.fetch.continue_request(
                request_id=request_id,
                url=url,
                method=method,
                post_data=post_data,
                headers=headers,
                intercept_response=intercept_response,
            )
        )

    async def fulfill_request(
        self,
        response_code: int,
        response_headers: typing.Optional[typing.List[HeaderEntry]] = None,
        binary_response_headers: typing.Optional[str] = None,
        body: typing.Optional[str] = None,
        response_phrase: typing.Optional[str] = None,
    ) -> None:
        request_id = (await self.response_future).request_id
        await self.tab.send(
            cdp.fetch.fulfill_request(
                request_id=request_id,
                response_code=response_code,
                response_headers=response_headers,
                binary_response_headers=binary_response_headers,
                body=body,
                response_phrase=response_phrase,
            )
        )

    async def continue_response(
        self,
        response_code: typing.Optional[int] = None,
        response_phrase: typing.Optional[str] = None,
        response_headers: typing.Optional[typing.List[HeaderEntry]] = None,
        binary_response_headers: typing.Optional[str] = None,
    ) -> None:
        request_id = (await self.response_future).request_id
        await self.tab.send(
            cdp.fetch.continue_response(
                request_id=request_id,
                response_code=response_code,
                response_phrase=response_phrase,
                response_headers=response_headers,
                binary_response_headers=binary_response_headers,
            )
        )



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Optional

from zendriver import cdp, util
from zendriver.core.element import Element

if TYPE_CHECKING:
    from zendriver.core.tab import Tab


logger = logging.getLogger(__name__)


async def cf_find_interactive_challenge(
    tab: Tab,
) -> tuple[Element | None, Element | None, Element | None]:
    """
    Finds the Cloudflare interactive challenge elements.

    This function scans the DOM for shadow roots and iframes that match the
    signature of a Cloudflare challenge.

    Returns:
        A tuple containing the host element, the shadow root element, and the
        challenge iframe element if found, otherwise (None, None, None).
    """
    logger.debug("Searching for Cloudflare interactive challenge elements...")
    doc = await tab.send(cdp.dom.get_document(-1, True))
    if not doc:
        logger.debug("DOM document not found.")
        return None, None, None

    # Find all nodes that have a "shadow root" (an isolated DOM).
    shadow_host_nodes = util.filter_recurse_all(
        doc, lambda n: hasattr(n, "shadow_roots") and bool(n.shadow_roots)
    )
    logger.debug(f"Found {len(shadow_host_nodes)} shadow host nodes.")

    # Iterate over each shadow DOM host node.
    for host_node in shadow_host_nodes:
        if not host_node.shadow_roots:
            continue

        # Iterate over each shadow root within the host.
        for shadow_root_node in host_node.shadow_roots:
            # Create an Element object for the shadow root.
            # The "tree" for a shadow root element is the shadow root node itself.
            shadow_root_element = Element(shadow_root_node, tab, shadow_root_node)

            # Check if the shadow root content is the Cloudflare challenge.
            html_content = await shadow_root_element.get_html()
            if "challenges.cloudflare.com" in html_content:
                logger.debug("Found Cloudflare challenge in a shadow root.")
                # If the shadow root is found, search for the specific iframe within it.
                for child_element in shadow_root_element.children:
                    if "challenges.cloudflare.com" in await child_element.get_html():
                        # Found! Create the host element and return everything.
                        logger.debug("Found challenge iframe.")
                        host_element = Element(host_node, tab, doc)
                        challenge_iframe = child_element
                        return host_element, shadow_root_element, challenge_iframe

    # If the loops finish without finding anything, return None.
    logger.debug("Cloudflare interactive challenge not found.")
    return None, None, None


async def cf_wait_for_interactive_challenge(
    tab: Tab, timeout: float = 5
) -> tuple[Element | None, Element | None, Element | None]:
    """
    Waits for the Cloudflare challenge iframe to appear and be visible.

    Args:
        timeout: The maximum time in seconds to wait for the elements.

    Returns:
        A tuple containing the host element, shadow root, and iframe element
        if found and visible, otherwise (None, None, None).
    """
    loop = asyncio.get_running_loop()
    start_time = loop.time()

    while loop.time() - start_time < timeout:
        logger.debug("Waiting for challenge elements to appear...")
        (
            host_element,
            shadow_root_element,
            challenge_iframe,
        ) = await cf_find_interactive_challenge(tab)
        if challenge_iframe and "display: none" not in challenge_iframe.attrs.get(
            "style", ""
        ):
            logger.debug("Cloudflare challenge elements found and visible.")
            return host_element, shadow_root_element, challenge_iframe
        await asyncio.sleep(0.5)

    logger.warning(
        f"Timeout: Cloudflare challenge elements not found or not visible within {timeout} seconds."
    )
    return None, None, None


async def cf_is_interactive_challenge_present(tab: Tab, timeout: float = 5) -> bool:
    """
    Checks if a Cloudflare interactive challenge is present and visible on the page.

    Args:
        timeout: The maximum time in seconds to wait for the challenge.

    Returns:
        True if the challenge is present and visible, False otherwise.
    """
    logger.debug(
        f"Checking for Cloudflare challenge with a timeout of {timeout} seconds."
    )
    _, _, challenge_iframe = await cf_wait_for_interactive_challenge(tab, timeout)
    is_present = challenge_iframe is not None
    logger.debug(f"Challenge present: {is_present}")
    return is_present


async def verify_cf(
    tab: Tab,
    click_delay: float = 5,
    timeout: float = 15,
    challenge_selector: Optional[str] = None,
    flash_corners: bool = False,
) -> None:
    """
    Finds and solves the Cloudflare checkbox challenge.

    The total time for finding and clicking is governed by `timeout`.

    Args:
        click_delay: The delay in seconds between clicks.
        timeout: The total time in seconds to wait for the challenge and solve it.
        challenge_selector: An optional CSS selector for the challenge input element.
        flash_corners: If True, flash the corners of the challenge element.

    Raises:
        TimeoutError: If the checkbox is not found or solved within the timeout.
    """
    logger.debug("Waiting for Cloudflare checkbox...")
    loop = asyncio.get_running_loop()
    start_time = loop.time()

    (
        host_element,
        shadow_root_element,
        challenge_iframe,
    ) = await cf_wait_for_interactive_challenge(tab, timeout)

    if not challenge_iframe:
        raise TimeoutError(
            f"Cloudflare checkbox not found or not visible within {timeout} seconds."
        )

    logger.debug("Cloudflare checkbox found, starting clicks.")

    await challenge_iframe.scroll_into_view()

    # To get the element's dimensions, its numeric 'node_id' is needed.
    # This ID is obtained from the underlying node object.
    logger.debug(
        f"Getting box model for challenge iframe (node_id: {challenge_iframe.node.node_id})"
    )
    box_model_result = await tab.send(
        cdp.dom.get_box_model(node_id=challenge_iframe.node.node_id)
    )
    # 'content_quad' is a list of 8 numbers representing the (x, y) coordinates
    # of the four corners of the element's "content-box": [x1, y1, x2, y2, x3, y3, x4, y4].
    content_quad = box_model_result.content

    # Extract only the x coordinates of the 4 corners
    x_coords = content_quad[0::2]
    # Extract only the y coordinates of the 4 corners
    y_coords = content_quad[1::2]

    min_x = min(x_coords)
    max_x = max(x_coords)
    min_y = min(y_coords)
    max_y = max(y_coords)

    click_x = min_x + (max_x - min_x) * 0.15
    click_y = min_y + (max_y - min_y) / 2

    logger.debug(
        f"Checkbox dimensions (content box): width={max_x - min_x}, height={max_y - min_y}"
    )

    if flash_corners:
        logger.debug("Showing flash_point at the 4 corners.")
        corners = list(zip(x_coords, y_coords))
        for x_corner, y_corner in corners:
            await tab.flash_point(x=x_corner, y=y_corner, duration=10)

    if not host_element:
        return

    input_element = None
    current_selector = None

    # selector priority
    # 1. challenge_selector
    # 2. input[name=cf-turnstile-response]
    # 3. input[name=cf_challenge_response]

    # sometimes turnstile challenge have inputs 2 and 3.
    # input 2 is default for turnstile challenge.

    current_selector = (
        challenge_selector
        if challenge_selector
        else "input[name=cf-turnstile-response]"
    )
    input_element = await host_element.query_selector(current_selector)

    if not input_element and not challenge_selector:
        current_selector = "input[name=cf_challenge_response]"
        input_element = await host_element.query_selector(current_selector)

    if not input_element:
        return

    checkbox_clicked = False

    async def check_input(
        input_el: Element, current_sltr: str, host_el: Element, ckbx_clckd: bool
    ) -> bool:
        """Checks if the input element is still present and without a value."""
        if not input_el:
            return False
        try:
            await input_el
            fresh_input = await host_el.query_selector(current_sltr)
        except Exception as e:
            raise Exception(f"Error checking input element: {e}.")
        if (input_el.attrs.get("value") or not fresh_input) and ckbx_clckd:
            # If the input disappears or gets a value, assume it's successfully completed.
            logger.debug("Input element check successful (disappeared or has value).")
            return False
        return True

    while await check_input(
        input_el=input_element,
        current_sltr=current_selector,
        host_el=host_element,
        ckbx_clckd=checkbox_clicked,
    ):
        if loop.time() - start_time >= timeout:
            raise TimeoutError(
                f"Could not solve the checkbox in {timeout} seconds (timeout during clicks)."
            )
        try:
            await tab.mouse_click(click_x, click_y)
            await asyncio.sleep(click_delay)
            checkbox_clicked = True
        except Exception as e:
            if "could not find position" in str(e) and checkbox_clicked:
                logger.debug("Checkbox disappeared after click. Assuming success.")
                break
            raise Exception(f"Error clicking checkbox: {e}.")

    logger.debug("Checkbox challenge completed. ✔")
    return



import asyncio

import zendriver as zd
from zendriver import cdp
from zendriver.cdp import runtime


async def main() -> None:
    browser = await zd.start()
    page = await browser.get(
        "https://cdpdriver.github.io/examples/console.html",
    )

    # Those 2 lines are equivalent and do the same thing
    await page.send(runtime.enable())
    await page.send(cdp.runtime.enable())

    await browser.stop()


if __name__ == "__main__":
    asyncio.run(main())
